@@ -25,7 +25,6 @@
 //!
 //! use sequoia_openpgp as openpgp;
 //! use openpgp::Cert;
-//! # use openpgp::crypto::hash::Digest;
 //! use openpgp::packet::signature;
 //! use openpgp::parse::Parse;
 //! use openpgp::policy::StandardPolicy;
@@ -77,7 +76,7 @@
 //! // Import the secret key material into gpg-agent.
 //! for k in cert.keys().secret() {
 //!     agent.import(P,
-//!                  &cert, k.parts_as_secret().expect("have secret"),
+//!                  &cert, k.key().parts_as_secret().expect("have secret"),
 //!                  true, true).await?;
 //! }
 //!
@@ -94,11 +93,11 @@
 //!     .keys().for_signing()
 //!     .next().expect("Have a signing-capable subkey.");
 //!
-//! let mut keypair = agent.keypair(&signing_key)?;
+//! let mut keypair = agent.keypair(signing_key.key())?;
 //!
 //! let mut signed_message = Vec::new();
 //! let message = Message::new(&mut signed_message);
-//! let message = Signer::new(message, keypair).build()?;
+//! let message = Signer::new(message, keypair)?.build()?;
 //! let mut message = LiteralWriter::new(message).build()?;
 //! message.write_all(b"I love you!")?;
 //! message.finalize()?;
@@ -138,7 +137,6 @@ use openpgp::{
         },
         SKESK,
     },
-    parse::Parse,
     policy::Policy,
     types::{
         HashAlgorithm,
@@ -300,6 +298,9 @@ pub struct Agent {
     c: assuan::Client,
     pinentry_mode: Option<PinentryMode>,
 
+    /// Agent version.
+    version: String,
+
     /// Whether this is an restricted connection.
     restricted: bool,
 }
@@ -436,6 +437,7 @@ impl Agent {
             c: connection,
             pinentry_mode: None,
             restricted: false,
+            version: "unknown".into(),
         };
 
         // Check whether this is an restricted connection.
@@ -443,9 +445,36 @@ impl Agent {
             agent.send_simple("GETINFO restricted").await.is_ok();
         agent.restricted = restricted;
 
+        if let Ok(v) = agent.get_agent_version().await {
+            agent.version = v;
+        }
+
         Ok(agent)
     }
 
+    /// Returns the keys managed by the agent.
+    async fn get_agent_version(&mut self) -> Result<String> {
+        self.send("GETINFO version")?;
+
+        let mut version = Vec::new();
+        while let Some(response) = self.next().await {
+            let response = response?;
+            match response {
+                Response::Ok { .. } => break,
+                Response::Comment { .. } => (),
+                Response::Data { partial } =>
+                    version.extend_from_slice(&partial[..]),
+                Response::Error { message, .. } =>
+                    return self.operation_failed(&message).await,
+                response =>
+                    return protocol_error(&response),
+            }
+        }
+
+        self.next().await; // Dummy read to send END.
+        Ok(String::from_utf8(version).map_err(|e| e.utf8_error())?)
+    }
+
     /// Returns whether this is an restricted connection.
     pub fn is_restricted(&self) -> bool {
         self.restricted
@@ -616,33 +645,6 @@ impl Agent {
                         unattended: bool,
                         overwrite: bool)
                         -> Result<bool>
-    {
-        // The gpg-agent shipped with GnuPG 2.4.x calculates the checksum
-        // over ECC artifacts differently.  Oddly, this seems to amount to
-        // adding 8 to the checksum.  See GnuPG commit
-        // 2b118516240b4bddd34c68c23a99bea56682a509.
-        use sequoia_openpgp::types::PublicKeyAlgorithm::*;
-        let mut r = self.import_int(
-            policy, cert, key, unattended, overwrite, 0).await;
-        if r.is_err() && (key.pk_algo() == ECDSA
-                          || key.pk_algo() == EdDSA
-                          || key.pk_algo() == ECDH)
-        {
-            r = self.import_int(
-                policy, cert, key, unattended, overwrite, 8).await;
-        }
-
-        r
-    }
-
-    async fn import_int(&mut self,
-                        policy: &dyn Policy,
-                        cert: &Cert,
-                        key: &Key<SecretParts, UnspecifiedRole>,
-                        unattended: bool,
-                        overwrite: bool,
-                        csum_offset: u16)
-                        -> Result<bool>
     {
         use ipc::sexp::*;
 
@@ -684,6 +686,42 @@ impl Agent {
                     .fold(0u16, |acc, v| acc.wrapping_add(*v as u16)));
         }
 
+        fn add_ecc_scalar(list: &mut Vec<Sexp>,
+                          mode: EccScalarChecksumMode,
+                          checksum: &mut u16,
+                          s: &mpi::ProtectedMPI)
+        {
+            match mode {
+                EccScalarChecksumMode::MPI =>
+                    addp(list, checksum, s),
+
+                EccScalarChecksumMode::SOS => {
+                    let mut v = s.value().to_vec();
+
+                    // If the high bit is set, we need to prepend a
+                    // zero byte, otherwise the agent will interpret
+                    // the value as signed, and thus negative.
+                    if *v.get(0).unwrap_or(&0) & 0x80 > 0 {
+                        v.insert(0, 0);
+                    }
+
+                    add_raw(list, "_", &v);
+
+                    // Compute the bit length.
+                    let nbits = v.len() as u32 * 8
+                        - s.value().get(0).map(|o| o.leading_zeros())
+                        .unwrap_or(0);
+
+                    // Compute the checksum.
+                    *checksum = checksum
+                        .wrapping_add(nbits as u16 >> 8)
+                        .wrapping_add(nbits as u16 & 0xff)
+                        .wrapping_add(v.iter().fold(
+                            0u16, |acc, v| acc.wrapping_add(*v as u16)));
+                },
+            }
+        }
+
         fn add_raw(list: &mut Vec<Sexp>, name: &str, data: &[u8]) {
             list.push(Sexp::String(name.into()));
             list.push(Sexp::String(data.into()));
@@ -795,10 +833,15 @@ impl Agent {
                     mpi::SecretKeyMaterial::DSA { x, }
                     | mpi::SecretKeyMaterial::ElGamal { x, } =>
                         addp(&mut skey, &mut checksum, x),
+
                     mpi::SecretKeyMaterial::EdDSA { scalar, }
                     | mpi::SecretKeyMaterial::ECDSA { scalar, }
-                    | mpi::SecretKeyMaterial::ECDH { scalar, } =>
-                        addp(&mut skey, &mut checksum, scalar),
+                    | mpi::SecretKeyMaterial::ECDH { scalar, } => {
+                        let mode = EccScalarChecksumMode::for_version(
+                            &self.version);
+                        add_ecc_scalar(&mut skey, mode, &mut checksum, scalar);
+                    },
+
                     mpi::SecretKeyMaterial::Unknown { mpis, rest, } => {
                         for m in mpis.iter() {
                             addp(&mut skey, &mut checksum, m);
@@ -823,7 +866,7 @@ impl Agent {
             transfer_key.push(s("curve", curve.to_string()));
         }
         transfer_key.push(Sexp::List(skey));
-        transfer_key.push(s("csum", checksum.wrapping_add(csum_offset)));
+        transfer_key.push(s("csum", checksum));
         transfer_key.push(protection);
 
         let transfer_key = Sexp::List(transfer_key);
@@ -1283,6 +1326,42 @@ impl Agent {
     }
 }
 
+/// ECC scalar checksum computation.
+///
+/// The gpg-agent shipped with GnuPG 2.4.x calculates the checksum
+/// over ECC scalars differently.  This is side-effect of the
+/// premature introduction of the simple-octet-string (SOS) encoding
+/// into the GnuPG code.
+///
+/// See GnuPG commit 2b118516240b4bddd34c68c23a99bea56682a509.
+enum EccScalarChecksumMode {
+    /// Sum over the MPI wire encoding in u16.
+    MPI,
+
+    /// Sum over the "opaque MPI encoding" and its bit length in u16.
+    ///
+    /// "Opaque MPI" is a gcrypt mode for MPIs that don't represent
+    /// integers, and mustn't be used in arithmetic functions.  Oddly
+    /// enough, if the left-most bit is set, gcrypt will prefix it
+    /// with a zero byte, presumably to prevent it from being
+    /// interpreted as a sign-bit.
+    SOS,
+}
+
+impl EccScalarChecksumMode {
+    fn for_version(v: &str) -> EccScalarChecksumMode {
+        if v.starts_with("2.0.")
+            || v.starts_with("2.1.")
+            || v.starts_with("2.2.")
+            || v.starts_with("2.3.")
+        {
+            EccScalarChecksumMode::MPI
+        } else {
+            EccScalarChecksumMode::SOS
+        }
+    }
+}
+
 /// Returns a convenient Err value for use in the state machines
 /// below.
 fn protocol_error<T>(response: &Response) -> Result<T> {
@@ -1870,11 +1949,11 @@ mod test {
     use openpgp::{
         Cert,
         crypto::{
-            hash::Digest,
             mpi::Ciphertext,
             Decryptor,
             Signer,
         },
+        parse::Parse,
         policy::StandardPolicy,
     };
 
@@ -1888,7 +1967,7 @@ mod test {
 
         for k in cert.keys().secret() {
             agent.import(P,
-                         &cert, k.parts_as_secret().expect("have secret"),
+                         &cert, k.key().parts_as_secret().expect("have secret"),
                          true, true).await?;
         }
 
@@ -1928,7 +2007,7 @@ mod test {
                 .primary_key().key().clone();
             let mut pair = KeyPair::new_for_gnupg_context(&ctx, &key)?;
             let algo = HashAlgorithm::default();
-            let digest = algo.context()?.into_digest()?;
+            let digest = algo.context()?.for_digest().into_digest()?;
             let _ = pair.sign(algo, &digest);
             Ok(())
         }
@@ -2021,7 +2100,7 @@ mod test {
         let key = testy.primary_key().key();
         let mut pair = KeyPair::new_for_gnupg_context(&ctx, &key)?;
         let algo = HashAlgorithm::default();
-        let digest = algo.context()?.into_digest()?;
+        let digest = algo.context()?.for_digest().into_digest()?;
         if let Err(err) = pair.sign(algo, &digest) {
             panic!("Signing: {}", err);
         }
@@ -2135,10 +2214,10 @@ mod test {
             let cert = Cert::from_bytes(crate::tests::key(file))?;
             import_keys(&mut agent, &cert).await?;
             let vc = cert.with_policy(P, None).expect("valid cert");
-            let key = vc.keys().for_signing().next().expect("have signing key");
-            let key_keygrip = Keygrip::of(key.mpis()).expect("has a keygrip");
+            let vka = vc.keys().for_signing().next().expect("have signing key");
+            let key_keygrip = Keygrip::of(vka.key().mpis()).expect("has a keygrip");
 
-            t!("Testing with {} ({})", key.fingerprint(), key_keygrip);
+            t!("Testing with {} ({})", vka.key().fingerprint(), key_keygrip);
             t!("gpg knows about:");
             let mut saw = false;
             for (i, info) in agent.list_keys().await.expect("works").iter().enumerate() {
@@ -2191,8 +2270,8 @@ mod test {
             }
 
             let algo = HashAlgorithm::default();
-            let digest = algo.context()?.into_digest()?;
-            let mut pair = KeyPair::new_for_gnupg_context(&ctx, &key)?
+            let digest = algo.context()?.for_digest().into_digest()?;
+            let mut pair = KeyPair::new_for_gnupg_context(&ctx, vka.key())?
                 // Don't prompt for the passphrase.
                 .suppress_pinentry();
             match pair.sign(algo, &digest) {
@@ -2223,8 +2302,8 @@ mod test {
             }
 
             let algo = HashAlgorithm::default();
-            let digest = algo.context()?.into_digest()?;
-            let mut pair = KeyPair::new_for_gnupg_context(&ctx, &key)?
+            let digest = algo.context()?.for_digest().into_digest()?;
+            let mut pair = KeyPair::new_for_gnupg_context(&ctx, vka.key())?
                 // Don't prompt for the passphrase.
                 .suppress_pinentry();
             match pair.sign(algo, &digest) {
@@ -2241,8 +2320,8 @@ mod test {
                 .expect("can forget");
 
             let algo = HashAlgorithm::default();
-            let digest = algo.context()?.into_digest()?;
-            let mut pair = KeyPair::new_for_gnupg_context(&ctx, &key)?
+            let digest = algo.context()?.for_digest().into_digest()?;
+            let mut pair = KeyPair::new_for_gnupg_context(&ctx, vka.key())?
                 // Don't prompt for the passphrase.
                 .suppress_pinentry();
             match pair.sign(algo, &digest) {
