@@ -20,16 +20,16 @@ pub fn best_effort_primary_uid(policy: &dyn Policy, cert: &Cert) -> String {
     // First, apply our policy.
     if let Ok(vcert) = cert.with_policy(policy, None) {
         if let Ok(primary) = vcert.primary_userid() {
-            primary_uid = Some(primary.value().to_vec());
+            primary_uid = Some(primary.userid().value().to_vec());
         }
     }
 
     // Second, apply the null policy.
     if primary_uid.is_none() {
-        let null = openpgp::policy::NullPolicy::new();
+        let null = unsafe { openpgp::policy::NullPolicy::new() };
         if let Ok(vcert) = cert.with_policy(&null, None) {
             if let Ok(primary) = vcert.primary_userid() {
-                primary_uid = Some(primary.value().to_vec());
+                primary_uid = Some(primary.userid().value().to_vec());
             }
         }
     }
@@ -37,7 +37,7 @@ pub fn best_effort_primary_uid(policy: &dyn Policy, cert: &Cert) -> String {
     // As a last resort, pick the first user id.
     if primary_uid.is_none() {
         if let Some(primary) = cert.userids().next() {
-            primary_uid = Some(primary.value().to_vec());
+            primary_uid = Some(primary.userid().value().to_vec());
         } else {
             // Special case, there is no user id.
             primary_uid = Some(b"(NO USER ID)"[..].into());
